#include <gl/gl.h>
#include <gl/device.h>
#include <stdio.h>
#include <sys/time.h>
#include "object.h"

/****
   This is a simple application to use the objects defined in\
   object.c. It draws some rotating objects, using
   different materials and different transformations.
   It allows for multiple views of the object in different windows.
   The right mouse button operates the pop-up menu.

   To add objects: 
      - create a new object in object.c (and object.h)
      - add a draw command in redraw
      - add a material and bind it
      - add a menu entry
   Up to 20 different objects are supported

   This is version 1.0.

   Written by Mark Overmars, Dept. of Computer Science, University of Utrecht,
   Email: markov@cs.ruu.nl
****/

#define MAXOBJ 20

int frames,back,wire,flat,incr; /* toggles to indicate whay of display */
int ob[MAXOBJ];                 /* indicates which objects to show */
int lcount;                     /* last frame cout */
float time_passed;                     /* the "time" so far */
float speed;			/* the rotation speed */
long win1, win2;                /* the window identifiers */

drw(obj,r1,r2,r3,t1,t2,t3,s1,s2,s3,mat)
/* draws a particular object. */
OBJECT obj;
float r1,r2,r3,t1,t2,t3,s1,s2,s3;
int mat;
{
    pushmatrix();
    rot(time_passed*r1,'X'); rot(time_passed*r2,'Y'); rot(time_passed*r3,'Z');
    translate(t1,t2,t3); scale(s1,s2,s3);
    lmbind(MATERIAL,mat); drawobject(obj,wire,flat);
    popmatrix();
}

redraw()
/* draws all the objects in the current window */
{
    char str[100];
    backface(back);
    /* clear the screen */
    zclear();
    RGBcolor(43,169,255);
    clear();
    /* do the drawing */
    if (ob[0]) drw(      sphere,  3.0,3.1,2.0,  2.2, 1.2, 2.0,  0.8,0.8,0.8, 1);
    if (ob[1]) drw(    cylinder,  3.2,2.6,1.8, -1.9, 1.2, 1.6,  0.3,0.3,1.0, 4);
    if (ob[2]) drw(        cube,  2.2,3.3,1.4, -1.0, 1.2,-1.5,  0.6,0.6,0.6, 3);
    if (ob[3]) drw(        cone,  4.2,2.1,3.2,  1.2, 1.3, 1.0,  0.5,0.5,0.8, 2);
    if (ob[3]) drw(      sphere,  4.2,2.1,3.2,  1.2, 1.3, 0.0,  0.5,0.5,0.5,51);
    if (ob[4]) drw(      sphere,  1.2,3.3,2.4, -2.0, 1.4,-2.1,  0.8,0.8,0.2, 5);
    if (ob[5]) drw(        cube,  3.2,3.6,2.4,  1.1, 1.6, 2.5,  0.8,0.3,0.1, 6);
    if (ob[6]) drw(       glass,  2.7,1.9,3.1,  1.6, 1.3, 0.7,  2.0,2.0,2.0, 7);
    if (ob[7]) drw(     pyramid,  2.3,2.7,3.3,  1.1, 2.3, 1.7,  0.6,0.6,0.5, 8);
    /* show frame count if required */
    if (frames && winget() == win1)
    {
    	cpack(0xffffff);
	cmov(-2.1,-2.1,4.0);
	sprintf(str,"%d frames per second",lcount);
	charstr(str);
    };
    /* ready */
    swapbuffers();
}

/* The identity matrix. */
Matrix idmat = {1.0,0.0,0.0,0.0,
	        0.0,1.0,0.0,0.0,
		0.0,0.0,1.0,0.0,
		0.0,0.0,0.0,1.0};

/* The different materials. */
float mater1[] = 
    {SPECULAR,0.8,0.0,0.0,DIFFUSE,0.4,0.0,0.0,SHININESS,40.0,LMNULL};
float mater51[] = 
    {SPECULAR,0.2,0.0,0.1,DIFFUSE,0.8,0.0,0.3,SHININESS,10.0,LMNULL};
float mater2[] =
    {SPECULAR,1.0,0.4,0.0,DIFFUSE,1.0,0.4,0.0,SHININESS,80.0,LMNULL};
float mater3[] =
    {SPECULAR,0.0,0.0,0.6,DIFFUSE,0.0,0.0,0.8,SHININESS,60.0,LMNULL};
float mater4[] =
    {SPECULAR,0.0,1.0,0.0,DIFFUSE,0.0,0.6,0.0,SHININESS,120.0,LMNULL};
float mater5[] =
    {SPECULAR,1.0,1.0,0.0,DIFFUSE,0.6,0.6,0.0,SHININESS,100.0,LMNULL};
float mater6[] =
    {SPECULAR,1.0,0.0,1.0,DIFFUSE,0.6,0.0,0.6,SHININESS,120.0,LMNULL};
float mater7[] =
    {SPECULAR,0.9,0.9,0.9,DIFFUSE,0.6,0.6,0.6,SHININESS,120.0,LMNULL};
float mater8[] =
    {SPECULAR,0.4,0.7,0.4,DIFFUSE,0.5,1.0,0.5,SHININESS,50.0,LMNULL};

/* The lightsources. */
float light1[] = {LCOLOR,1.0,1.0,1.0,POSITION,10.0,10.0,5.0,0.0,LMNULL};
float light2[] = {LCOLOR,1.0,1.0,1.0,POSITION,-10.0,10.0,5.0,0.0,LMNULL};

/* The lightmodel. */
float model[] = {AMBIENT,0.4,0.4,0.4,LMNULL};

init()
/* Initializes all settings for a window. */
{
    zbuffer(TRUE);
    doublebuffer();
    RGBmode();
    gconfig();
    mmode(MVIEWING);
    perspective(400,1.0,1.0,34.0);
    loadmatrix(idmat);
    /* set up the devices */
    qdevice(MOUSE1);
    qdevice(MOUSE3);
    qdevice(ESCKEY);
    qdevice(WINFREEZE);
    qdevice(WINTHAW);
    /* define materials and lights */
    lmdef(DEFMATERIAL, 1, 11, mater1);
    lmdef(DEFMATERIAL,51, 11, mater51);
    lmdef(DEFMATERIAL, 2, 11, mater2);
    lmdef(DEFMATERIAL, 3, 11, mater3);
    lmdef(DEFMATERIAL, 4, 11, mater4);
    lmdef(DEFMATERIAL, 5, 11, mater5);
    lmdef(DEFMATERIAL, 6, 11, mater6);
    lmdef(DEFMATERIAL, 7, 11, mater7);
    lmdef(DEFMATERIAL, 8, 11, mater8);
    lmdef(DEFLIGHT, 1, 10, light1);
    lmdef(DEFLIGHT, 2, 10, light2);
    lmdef(DEFLMODEL, 1, 5, model);
    lmbind(LIGHT0,1);
    lmbind(LIGHT1,2);
    lmbind(LMODEL,1);
}

/*** Handling the menu. ***/

long mainmenu, confirmmenu, speedmenu, formmenu, lightmenu, extramenu, omenu;

createobjectmenu()
{
    omenu = newpup();
    addtopup(omenu,"  Show All %x71 | Show None %x72 %l");
    if (ob[0]) {addtopup(omenu,"-> Sphere %x41");}
         else  {addtopup(omenu,"   Sphere %x41");};
    if (ob[1]) {addtopup(omenu,"-> Cylinder %x42");}
         else  {addtopup(omenu,"   Cylinder %x42");};
    if (ob[2]) {addtopup(omenu,"-> Cube %x43");}
         else  {addtopup(omenu,"   Cube %x43");};
    if (ob[3]) {addtopup(omenu,"-> Ice Cream %x44");}
         else  {addtopup(omenu,"   Ice Cream %x44");};
    if (ob[4]) {addtopup(omenu,"-> Disk %x45");}
         else  {addtopup(omenu,"   Disk %x45");};
    if (ob[5]) {addtopup(omenu,"-> Diamond %x46");}
         else  {addtopup(omenu,"   Diamond %x46");};
    if (ob[6]) {addtopup(omenu,"-> Glass %x47");}
         else  {addtopup(omenu,"   Glass %x47");};
    if (ob[7]) {addtopup(omenu,"-> Pyramid %x48");}
         else  {addtopup(omenu,"   Pyramid %x48");};
}

createformmenu()
{
    formmenu = newpup();
    if (wire)  {addtopup(formmenu,"-> Wire Frame %x11");}
         else  {addtopup(formmenu,"   Wire Frame %x11");};
    if (wire)  {addtopup(formmenu,"   Filled %x12");}
         else  {addtopup(formmenu,"-> Filled %x12");};
}

createlightmenu()
{
    lightmenu = newpup();
    if (flat)  {addtopup(lightmenu,"-> Flat Shaded %x21");}
         else  {addtopup(lightmenu,"   Flat Shaded %x21");};
    if (flat)  {addtopup(lightmenu,"   Gouraud Shaded %x22");}
         else  {addtopup(lightmenu,"-> Gouraud Shaded %x22");};
}

createextramenu()
{
    extramenu = newpup();
    if (back)  {addtopup(extramenu,"-> Backface Culling %x31");}
         else  {addtopup(extramenu,"   Backface Culling %x31");};
    if (frames){addtopup(extramenu,"-> Show Frames %x32 %l");}
         else  {addtopup(extramenu,"   Show Frames %x32 %l");};
    addtopup(extramenu,"  Faster %x33 |  Slower %x34");
}

createmenu()
/* Creates the menu */
{
    createobjectmenu();
    createformmenu();
    createlightmenu();
    createextramenu();
    mainmenu = newpup();
    addtopup(mainmenu,"Objects Demo %t");
    if (win2 == 0){addtopup(mainmenu,"Show Front View %x1 %l");}
    addtopup(mainmenu,"Objects %m",omenu);
    addtopup(mainmenu,"Display Mode %m",formmenu);
    addtopup(mainmenu,"Lighting Mode %m",lightmenu);
    addtopup(mainmenu,"Extras %m %l",extramenu);
    addtopup(mainmenu,"Exit %x5");
    confirmmenu = newpup();
    addtopup(confirmmenu,"Are You Sure? %t|Yes|No");
}

destroymenu()
/* kills all the menus */
{
    freepup(mainmenu);
    freepup(lightmenu);
    freepup(formmenu);
    freepup(confirmmenu);
    freepup(omenu);
    freepup(extramenu);
}


postmenu()
/* Displays the menu and takes action. */
{
    long entry;
    int i;
    entry = dopup(mainmenu);
    switch (entry) 
    {
	case 1:	setpup(mainmenu,1,PUP_GREY);
    		prefposition(600,1000,500,900);
    		win2 = winopen("Front View");
    		winconstraints();
		iconsize(85,66); keepaspect(1,1);
		winconstraints();
    		init();
    		lookat(10.0,0.0,0.0,0.0,0.0,0.0,0.0);
		break;
	case 5:	if (dopup(confirmmenu) == 1) exit(0);
		break;
	case 11: wire = 1; break;
	case 12: wire = 0; break;
	case 21: flat = 1; break;
	case 22: flat = 0; break;
	case 31: back = 1 - back; break;
	case 32: frames = 1 - frames; break;
	case 33: speed = speed * 1.5; break;
	case 34: speed = speed / 1.5; break;
	case 41:
	case 42:
	case 43:
	case 44:
	case 45:
	case 46:
	case 47:
	case 48:
	case 49:
	case 50:
	case 51:
	case 52:
	case 53:
	case 54:
	case 55:
	case 56:
	case 57:
	case 58:
	case 59:
	case 60: ob[entry-41] = 1 - ob[entry-41]; break;
	case 71: for (i=0; i<MAXOBJ; i++) ob[i] = 1; break;
	case 72: for (i=0; i<MAXOBJ; i++) ob[i] = 0; break;
    };
    destroymenu();
    createmenu();
}

main()
{
    Device dev;
    short val;
    struct timeval tp;
    struct timezone tzp;
    long passed,lastsec,lasttime;
    int count;
    int i;
    /* start up the stuff */
    makeobjects();
    prefposition(100,500,500,900);
    win1 = winopen("Objects");
    winconstraints(); iconsize(85,66); keepaspect(1,1); winconstraints();
    init();
    lookat(0.0,0.0,10.0,0.0,0.0,0.0,0.0);
    win2 = 0;
    for (i=0; i<MAXOBJ; i++) ob[i] = 0;
    ob[0] = ob[1] = ob[2] = 1;
    frames = 1; back = 1; wire = 0; flat = 0; incr = 1; time_passed= 0.0;
    speed = 5.0;
    createmenu();
    /* ready to go */
    gettimeofday(&tp,&tzp);
    lastsec = tp.tv_sec;
    lasttime = tp.tv_usec;
    while (1)
    {
	/* draw the stuff */
	gettimeofday(&tp,&tzp);
        if (incr)
  	{
	    passed = 1000000 * (tp.tv_sec - lastsec) + (tp.tv_usec - lasttime);
	    time_passed = time_passed + speed * passed/250000.0;
	};
        if (lastsec != tp.tv_sec) {lcount = count; count = 0;};
        count++;
        lastsec = tp.tv_sec;
        lasttime = tp.tv_usec;
	winset(win1); redraw();
	if (win2) {winset(win2); redraw();}
	/* listen at the devices */
	if (qtest())
	{
	    dev = qread(&val);
	    if (dev == MOUSE1 && val )
            {
		postmenu();
		gettimeofday(&tp,&tzp);
        	lastsec = tp.tv_sec;
        	lasttime = tp.tv_usec;
            }
	    else if (dev == ESCKEY) { exit(0); }
	    else if (dev == REDRAW)
	        { winset(val); reshapeviewport();}
	    else if (dev == REDRAWICONIC)
	        { winset(val); reshapeviewport();}
	    else if (dev == WINTHAW)
	        { winset(val); reshapeviewport();}
	    else if (dev == MOUSE3 && val){ incr = 1 - incr;}
        }
    }
}
