#include <gl/gl.h>
#include <math.h>
#include <stdio.h>
#include "object.h"
/*
   This library contains a number of basic objects plus a routine to
   draw them. All objects have their center at point (0,0,0) and are
   maximal in a box of size 2*2*2 around the origin. The currently
   supported objects are:

   cube           	: a simple cube, origin is the center.
   sphere         	: a nice looking sphere, origin is the center.
   cylinder       	: a cylinder.
   cone                 : a cone
   glass		: a nice glass
   pyramid              : guess what



   The following routines exist for use:

   makeobjects()
		should be called once before using the objects.

   drawobject(obj,wire,flat)
		to draw object obj. wire indicate whether wire frame
		drawing is required. flat indicate whether flat shading
		should be used.

*/

/*************************
 Making Spin Objects
*************************/

#define PI    3.14159265
#define ZERO  0.00001

makespinobject(obj,smooth,rot,n,x1,z1,nx1,nz1,x2,z2,nx2,nz2)
/* makes a spin object */
OBJECT *obj;         /* the object to be created */
int smooth;          /* whether smooth or not */
int rot;             /* the number of rotation steps */
int n;               /* the number of segments to spin */
float x1[],z1[];     /* (x,z) coordinates of first endpoint of segments */
float nx1[],nz1[];   /* their normals */
float x2[],z2[];     /* (x,z) coordinates of second endpoint of segments */
float nx2[],nz2[];   /* their normals */
{
    float th,dth,a1,a2;
    PATCH *p;
    int i,j;
    obj->numb = rot*n;
    p = obj->pat = (PATCH *) malloc(obj->numb * sizeof(PATCH));
    dth = 2.0*PI/rot;
    for (i=0; i<n; i++)
    {
        for (j=0; j<rot; j++)
        {
 	    th =  dth * j;   
	    if (smooth) {a1 = th; a2 =th+dth;} else { a1 = a2 = th +dth/2.0;};
	    if (x1[i] < ZERO || x2[i] <ZERO) {p->numb=3;} else {p->numb = 4;};
	    p->v[0].x = x1[i]*sin(th);     p->n[0].x = nx1[i]*sin(a1);
	    p->v[0].y = x1[i]*cos(th);     p->n[0].y = nx1[i]*cos(a1);
	    p->v[0].z = z1[i];             p->n[0].z = nz1[i];
	    p->v[1].x = x1[i]*sin(th+dth); p->n[1].x = nx1[i]*sin(a2);
	    p->v[1].y = x1[i]*cos(th+dth); p->n[1].y = nx1[i]*cos(a2);
	    p->v[1].z = z1[i];             p->n[1].z = nz1[i];
	    p->v[2].x = x2[i]*sin(th+dth); p->n[2].x = nx2[i]*sin(a2);
	    p->v[2].y = x2[i]*cos(th+dth); p->n[2].y = nx2[i]*cos(a2);
	    p->v[2].z = z2[i];             p->n[2].z = nz2[i];
	    p->v[3].x = x2[i]*sin(th);     p->n[3].x = nx2[i]*sin(a1);
	    p->v[3].y = x2[i]*cos(th);     p->n[3].y = nx2[i]*cos(a1);
	    p->v[3].z = z2[i];             p->n[3].z = nz2[i];
	    if (x1[i] <ZERO) 
	    {
		p->v[1] = p->v[2]; p->n[1] = p->n[2];
		p->v[2] = p->v[3]; p->n[2] = p->n[3];
	    };
	    p++;
	}
    }
}

makesmoothspinobject(obj,rot,n,x,z)
/* makes a smooth, connected spin object */
OBJECT *obj;         /* the object to be created */
int rot;             /* the number of rotation steps */
int n;               /* the number of points */
float x[],z[];       /* (x,z) coordinates of endpoints of segments */
{
    float x1[20],z1[20],nx1[20],nz1[20],x2[20],z2[20],nx2[20],nz2[20],l[20];
    int i;
    /* fill in the points */
    x1[0] = x[0]; z1[0] = z[0];
    for (i=1; i<n ; i++)
    	{ x1[i] = x2[i-1] = x[i]; z1[i] = z2[i-1] = z[i]; };
    /* fill in the normals */
    for (i=0; i<n-1 ; i++)
      {l[i] = sqrt((x[i+1]-x[i])*(x[i+1]-x[i]) + (z[i+1]-z[i])*(z[i+1]-z[i]));};
    nx1[0]   = 0.0; nz1[0]   =  1.0;
    nx2[n-2] = 0.0; nz2[n-2] = -1.0;
    for (i=1; i<n-1 ; i++)
    {
	nx1[i] = nx2[i-1] = ((z[i-1]-z[i])/l[i-1] + (z[i] - z[i+1])/l[i] )/2.0;
	nz1[i] = nz2[i-1] = ((x[i]-x[i-1])/l[i-1] + (x[i+1] - x[i])/l[i] )/2.0;
    }
    /* make it */
    makespinobject(obj,1,rot,n-1,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

/*Making some nice objects */

float x1[20],z1[20],nx1[20],nz1[20],x2[20],z2[20],nx2[20],nz2[20];

makesphere(sphere,n)
OBJECT *sphere;
int n;
{
    int i;
    x1[0] = nx1[0] = 0.0; z1[0] = nz1[0] = 1.0;
    for (i=0; i<n-1; i++)
    {
	x2[i] = nx2[i] = x1[i+1] = nx1[i+1] = sin((PI/n)*(1.0+i));
	z2[i] = nz2[i] = z1[i+1] = nz1[i+1] = cos((PI/n)*(1.0+i));
    }
    x2[n-1] = nx2[n-1] = 0.0; z2[n-1] = nz2[n-1] = -1.0;
    makespinobject(sphere,1,2*n,n,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

makecylinder(cylinder,n)
OBJECT *cylinder;
int n;
{
    x1[0] =  0.0; nx1[0] = 0.0; z1[0] =  1.0; nz1[0] =  1.0;
    x2[0] =  1.0; nx2[0] = 0.0; z2[0] =  1.0; nz2[0] =  1.0;
    x1[1] =  1.0; nx1[1] = 1.0; z1[1] =  1.0; nz1[1] =  0.0;
    x2[1] =  1.0; nx2[1] = 1.0; z2[1] = -1.0; nz2[1] =  0.0;
    x1[2] =  1.0; nx1[2] = 0.0; z1[2] = -1.0; nz1[2] = -1.0;
    x2[2] =  0.0; nx2[2] = 0.0; z2[2] = -1.0; nz2[2] = -1.0;
    makespinobject(cylinder,1,2*n,3,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

makecone(cone,n)
OBJECT *cone;
int n;
{
    x1[0] =  0.0; nx1[0] = 2.0/sqrt(5.0); z1[0] =  1.0; nz1[0] = 1.0/sqrt(5.0);
    x2[0] =  1.0; nx2[0] = 2.0/sqrt(5.0); z2[0] = -1.0; nz2[0] = 1.0/sqrt(5.0);
    x1[1] =  1.0; nx1[1] = 0.0;           z1[1] = -1.0; nz1[1] = -1.0;
    x2[1] =  0.0; nx2[1] = 0.0;           z2[1] = -1.0; nz2[1] = -1.0;
    makespinobject(cone,1,2*n,2,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

makecube(cube)
OBJECT *cube;
{
    x1[0] =  0.0;       nx1[0] = 0.0; z1[0] =  1.0; nz1[0] =  1.0;
    x2[0] =  sqrt(2.0); nx2[0] = 0.0; z2[0] =  1.0; nz2[0] =  1.0;
    x1[1] =  sqrt(2.0); nx1[1] = 1.0; z1[1] =  1.0; nz1[1] =  0.0;
    x2[1] =  sqrt(2.0); nx2[1] = 1.0; z2[1] = -1.0; nz2[1] =  0.0;
    x1[2] =  sqrt(2.0); nx1[2] = 0.0; z1[2] = -1.0; nz1[2] = -1.0;
    x2[2] =  0.0;       nx2[2] = 0.0; z2[2] = -1.0; nz2[2] = -1.0;
    makespinobject(cube,0,4,3,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}

makepyramid(pyramid)
OBJECT *pyramid;
{
    x1[0] = 0.0;      nx1[0] = 2.0/sqrt(5.0);z1[0] =  1.0;nz1[0] =1.0/sqrt(5.0);
    x2[0] = sqrt(2.0);nx2[0] = 2.0/sqrt(5.0);z2[0] = -1.0;nz2[0] =1.0/sqrt(5.0);
    x1[1] = sqrt(2.0);nx1[1] = 0.0;          z1[1] = -1.0;nz1[1] = -1.0;
    x2[1] = 0.0;      nx2[1] = 0.0;          z2[1] = -1.0;nz2[1] = -1.0;
    makespinobject(pyramid,0,4,2,x1,z1,nx1,nz1,x2,z2,nx2,nz2);
}


makeglass(glass,n)
OBJECT *glass;
int n;
{
    x1[ 0] = 0.000; z1[ 0] = 0.575;
    x1[ 1] = 0.050; z1[ 1] = 0.575;
    x1[ 2] = 0.100; z1[ 2] = 0.600;
    x1[ 3] = 0.200; z1[ 3] = 0.700;
    x1[ 4] = 0.250; z1[ 4] = 0.800;
    x1[ 5] = 0.250; z1[ 5] = 1.000;
    x1[ 6] = 0.275; z1[ 6] = 1.000;
    x1[ 7] = 0.275; z1[ 7] = 0.800;
    x1[ 8] = 0.250; z1[ 8] = 0.700;
    x1[ 9] = 0.200; z1[ 9] = 0.625;
    x1[10] = 0.100; z1[10] = 0.550;
    x1[11] = 0.050; z1[11] = 0.500;
    x1[12] = 0.025; z1[12] = 0.450;
    x1[13] = 0.025; z1[13] = 0.100;
    x1[14] = 0.050; z1[14] = 0.050;
    x1[15] = 0.250; z1[15] = 0.025;
    x1[16] = 0.250; z1[16] = 0.000;
    x1[17] = 0.225; z1[17] = 0.000;
    x1[18] = 0.000; z1[18] = 0.000;
    makesmoothspinobject(glass,n,19,x1,z1);
}

/*************************************************************

  WHAT THE USER SEES

*************************************************************/


makeobjects()
/* Creates all the objects. */
{
    makecube(&cube);
    makesphere(&sphere,12);
    makecylinder(&cylinder,12);
    makecone(&cone,12);
    makeglass(&glass,18);
    makepyramid(&pyramid);
}
    
drawobject(obj,wire,flat)
/* Draws a defined object. wire indicates whether wireframe. 
   flat indicates whether flat shaded. */
OBJECT obj;
int wire,flat;
{
    int i,j;
    PATCH *p;
    p = obj.pat;
    if (flat) {shademodel(FLAT);} else {shademodel(GOURAUD);};
    for (i=0; i < obj.numb ; i++)
    {
	if (wire) {bgnclosedline();} else {bgnpolygon();};
	for (j=0; j < p->numb; j++)
	    { n3f(&p->n[j]); v3f(&p->v[j]);};
	if (wire) {endclosedline();} else {endpolygon();};
	p++;
    }
}

